@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2010 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: CmdExec.cc,v 1.147 2009/07/17 12:48:41 lav Exp $ */
+/* $Id: CmdExec.cc,v 1.153 2011/04/29 04:58:27 lav Exp $ */
 
 #include <config.h>
 #include <stddef.h>
@@ -237,24 +237,25 @@ restart:
 	 }
 	 return;
       }
-      if(new_job)
-      {
-	 if(new_job->jobno<0)
-	    new_job->AllocJobno();
+      RevertToSavedSession();
+      if(new_job) {
 	 if(!new_job->cmdline)
 	    new_job->cmdline.set_allocated(cmdline.borrow());
-	 new_job->SetParentFg(this,!background);
+	 AddNewJob(new_job);
       }
-      AddWaiting(new_job);
-      if(background)
-      {
-	 if(new_job)
-	 {
-	    Roll(new_job);
-	    if(!new_job->Done())
-	       SuspendJob(new_job);
-	 }
-      } // background
+   }
+}
+
+void CmdExec::AddNewJob(Job *new_job)
+{
+   if(new_job->jobno<0)
+      new_job->AllocJobno();
+   new_job->SetParentFg(this,!background);
+   AddWaiting(new_job);
+   if(background) {
+      Roll(new_job);
+      if(!new_job->Done())
+	 SuspendJob(new_job);
    }
 }
 
@@ -401,11 +402,7 @@ int CmdExec::Do()
 		  session->Close();
 		  exit_code=0;
 		  builtin=BUILTIN_NONE;
-		  char *cmd=string_alloca(6+3+2*strlen(loc));
-		  strcpy(cmd,"open \"");
-		  unquote(cmd+strlen(cmd),loc);
-		  strcat(cmd,"\";");
-		  PrependCmd(cmd);
+		  PrependCmd(xstring::get_tmp("open ").append_quoted(loc));
 	       	  return MOVED;
 	       }
 	    }
@@ -705,61 +702,52 @@ void CmdExec::ShowRunStatus(const SMTaskRef<StatusLine>& s)
    }
 }
 
-void CmdExec::PrintStatus(int v,const char *prefix)
+xstring& CmdExec::FormatStatus(xstring& s,int v,const char *prefix)
 {
-   SessionJob::PrintStatus(v,prefix);
+   SessionJob::FormatStatus(s,v,prefix);
    if(builtin)
    {
-      xstring_ca s(args->Combine());
-      printf(_("\tExecuting builtin `%s' [%s]\n"),s.get(),session->CurrentStatus());
-      return;
+      xstring_ca ac(args->Combine());
+      return s.appendf(_("\tExecuting builtin `%s' [%s]\n"),ac.get(),session->CurrentStatus());
    }
    if(queue_feeder)
    {
       if(IsSuspended())
-	 printf("%s%s\n",prefix,_("Queue is stopped."));
+	 s.appendf("%s%s\n",prefix,_("Queue is stopped."));
       BuryDoneJobs();
       for(int i=0; i<waiting_num; i++)
       {
 	 if(i==0)
-	    printf("%s%s ",prefix,_("Now executing:"));
+	    s.appendf("%s%s ",prefix,_("Now executing:"));
 	 if(v==0)
-	    waiting[i]->ListOneJob(v);
+	    waiting[i]->FormatOneJob(s,v);
 	 else
-	    waiting[i]->PrintJobTitle();
+	    waiting[i]->FormatJobTitle(s);
 	 if(i+1<waiting_num)
-	    printf("%s\t-",prefix);
+	    s.appendf("%s\t-",prefix);
       }
-      queue_feeder->PrintStatus(v,prefix);
-      return;
+      return queue_feeder->FormatStatus(s,v,prefix);
    }
    if(waiting_num==1)
-   {
-      printf(_("\tWaiting for job [%d] to terminate\n"),waiting[0]->jobno);
-      return;
-   }
+      return s.appendf(_("\tWaiting for job [%d] to terminate\n"),waiting[0]->jobno);
    else if(waiting_num>1)
    {
-      printf(_("\tWaiting for termination of jobs: "));
+      s.appendf(_("\tWaiting for termination of jobs: "));
       for(int i=0; i<waiting_num; i++)
       {
-	 printf("[%d]",waiting[i]->jobno);
-	 printf("%c",i+1<waiting_num?' ':'\n');
+	 s.appendf("[%d]",waiting[i]->jobno);
+	 s.append(i+1<waiting_num?' ':'\n');
       }
-      return;
+      return s;
    }
    if(cmd_buf.Size()>0)
-   {
-      // xgettext:c-format
-      printf(_("\tRunning\n"));
-   }
+      s.append(_("\tRunning\n"));
    else if(feeder)
-   {
-      printf(_("\tWaiting for command\n"));
-   }
+      s.append(_("\tWaiting for command\n"));
+   return s;
 }
 
-CmdExec::CmdExec(FileAccess *f,LocalDirectory *c) : SessionJob(f?f:new DummyProto)
+void CmdExec::init(LocalDirectory *c)
 {
    // add this to chain
    next=chain;
@@ -805,7 +793,16 @@ CmdExec::CmdExec(FileAccess *f,LocalDirectory *c) : SessionJob(f?f:new DummyProt
 
    Reconfig();
 }
-
+CmdExec::CmdExec(FileAccess *s,LocalDirectory *c)
+   : SessionJob(s?s:new DummyProto)
+{
+   init(c);
+}
+CmdExec::CmdExec(const CmdExec *parent)
+   : SessionJob(parent->session->Clone())
+{
+   init(parent->cwd->Clone());
+}
 CmdExec::~CmdExec()
 {
    // remove this from chain.
@@ -831,11 +828,9 @@ const char *CmdExec::FormatPrompt(const char *scan)
    {
       const char *home=session->GetHome();
       if(home && strcmp(home,"/") && !strncmp(cwd,home,strlen(home))
-	    && (cwd[strlen(home)]=='/' || cwd[strlen(home)]==0))
+      && (cwd[strlen(home)]=='/' || cwd[strlen(home)]==0))
       {
-	 char *cwdbuf=string_alloca(strlen(cwd)-strlen(home)+2);
-	 sprintf(cwdbuf,"~%s",cwd+strlen(home));
-	 cwd=cwdbuf;
+	 cwd=xstring::format("~%s",cwd+strlen(home));
       }
    }
    const char *cwdb=session->GetCwd();
@@ -1015,47 +1010,39 @@ void CmdExec::SetInteractive(bool i)
    interactive=i;
 }
 
-int CmdExec::unquote(char *buf,const char *str)
+xstring& xstring::append_quoted(const char *str,int len)
 {
-   char *buf0=buf;
-   while(*str)
+   if(!CmdExec::needs_quotation(str,len))
+      return append(str);
+
+   append('"');
+   while(len>0)
    {
       if(*str=='"' || *str=='\\')
-	 *buf++='\\';
-      *buf++=*str++;
+	 append('\\');
+      append(*str++);
+      len--;
    }
-   *buf=0;
-   return buf-buf0;
+   return append('"');
 }
 
-const char *CmdExec::unquote(const char *str)
+bool CmdExec::needs_quotation(const char *buf,int len)
 {
-   static xstring ret;
-   ret.get_space(strlen(str)*2);
-   ret.set_length(unquote(ret.get_non_const(),str));
-   return ret;
-}
-
-bool CmdExec::needs_quotation(const char *buf)
-{
-   while(*buf)
+   while(len>0)
    {
       if(*buf==' ' || *buf=='\t')
 	 return true;
       if(strchr("\"'\\&|>;",*buf))
 	 return true;
       buf++;
+      len--;
    }
    return false;
 }
 
 void CmdExec::FeedQuoted(const char *c)
 {
-   char *buf=(char*)alloca(strlen(c)*2+2+1);
-   buf[0]='"';
-   unquote(buf+1,c);
-   strcat(buf,"\"");
-   FeedCmd(buf);
+   FeedCmd(xstring::get_tmp("").append_quoted(c));
 }
 
 // implementation is here because it depends on CmdExec.
@@ -1067,10 +1054,7 @@ char *ArgV::CombineQuoted(int start) const
    for(;;)
    {
       const char *arg=String(start++);
-      if(CmdExec::needs_quotation(arg))
-	 res.vappend("\"",CmdExec::unquote(arg),"\"",NULL);
-      else
-	 res.append(arg);
+      res.append_quoted(arg);
       if(start>=Count())
 	 return(res.borrow());
       res.append(' ');
@@ -1202,8 +1186,7 @@ Job *CmdExec::default_cmd()
 {
    const char *op=args->a0();
 #ifdef WITH_MODULES
-   char *modname=(char*)alloca(4+strlen(op)+1);
-   sprintf(modname,"cmd-%s",op);
+   const char *modname=xstring::cat("cmd-",op,NULL);
    if(module_load(modname,0,0)==0)
    {
       eprintf("%s\n",module_error_message());
@@ -1216,6 +1199,20 @@ Job *CmdExec::default_cmd()
    return 0;
 #endif
 }
+Job *CmdExec::builtin_local()
+{
+   saved_session=session.borrow();
+   session=FileAccess::New("file");
+   if(!session) {
+      eprintf("%s: cannot create local session\n",args->a0());
+      RevertToSavedSession();
+      return 0;
+   }
+   session->SetCwd(cwd->GetName());
+   args->delarg(0);
+   builtin=BUILTIN_EXEC_RESTART;
+   return this;
+}
 
 void CmdExec::FeedArgV(const ArgV *args,int start)
 {
