@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ResMgr.h,v 1.55 2008/11/27 05:56:26 lav Exp $ */
+/* $Id: ResMgr.h,v 1.56 2011/02/10 13:28:56 lav Exp $ */
 
 #ifndef RESMGR_H
 #define RESMGR_H
@@ -108,10 +108,11 @@ public:
    static const char *IPv6AddrValidate(xstring_c *value);
 #endif
    static const char *UNumberPairValidate(xstring_c *value);
-   static const char *FileAccessible(xstring_c *value,int mode,int want_dir=0);
+   static const char *FileAccessible(xstring_c *value,int mode,bool want_dir=false);
    static const char *FileReadable(xstring_c *value);
    static const char *FileExecutable(xstring_c *value);
    static const char *DirReadable(xstring_c *value);
+   static const char *FileCreatable(xstring_c *value);
    static const char *CharsetValidate(xstring_c *value);
    static const char *NoClosure(xstring_c *);
    static bool str2bool(const char *value);
