@@ -1,11 +1,11 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1998-2008 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1998-2010 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: SleepJob.cc,v 1.38 2008/08/05 12:10:12 lav Exp $ */
+/* $Id: SleepJob.cc,v 1.45 2011/04/29 04:58:28 lav Exp $ */
 
 #include <config.h>
 #include <ctype.h>
@@ -47,14 +47,15 @@ SleepJob::~SleepJob()
 
 int SleepJob::Do()
 {
+   int m=STALL;
    if(Done())
-      return STALL;
+      return m;
 
    if(waiting.count()>0)
    {
       Job *j=FindDoneAwaitedJob();
       if(!j)
-	 return STALL;
+	 return m;
       exit_code=j->ExitCode();
       if(!repeat || (++repeat_count>=max_repeat_count && max_repeat_count)
       || exit_code==break_code || (continue_code!=-1 && exit_code!=continue_code))
@@ -68,6 +69,7 @@ int SleepJob::Do()
       Reset();
       exec=(CmdExec*)j; // we are sure it is CmdExec.
       RemoveWaiting(j);
+      m=MOVED;
    }
 
    if(Stopped())
@@ -77,7 +79,6 @@ int SleepJob::Do()
 	 if(!exec)
 	 {
 	    exec=new CmdExec(session.borrow(),saved_cwd.borrow());
-	    exec->SetParentFg(this);
 	    exec->AllocJobno();
 	    exec->cmdline.vset("(",cmd.get(),")",NULL);
 	 }
@@ -89,15 +90,38 @@ int SleepJob::Do()
       done=true;
       return MOVED;
    }
-   return STALL;
+   return m;
 }
 
-void SleepJob::PrintStatus(int,const char *prefix)
+const char *SleepJob::Status()
+{
+   if(Stopped() || TimeLeft().Seconds()<=1)
+      return "";
+
+   if(IsInfty())
+      return(_("Sleeping forever"));
+   return xstring::cat(_("Sleep time left: "),
+      TimeLeft().toString(TimeInterval::TO_STR_TRANSLATE),
+      NULL);
+}
+
+xstring& SleepJob::FormatStatus(xstring& buf,int,const char *prefix)
 {
    if(repeat)
+      buf.appendf(_("\tRepeat count: %d\n"),repeat_count);
+   const char *s=Status();
+   if(s[0])
+      buf.appendf("\t%s\n",s);
+   return buf;
+}
+void SleepJob::ShowRunStatus(const SMTaskRef<StatusLine>& s)
+{
+   if(Stopped())
+      Job::ShowRunStatus(s);
+   else
    {
-      printf(_("\tRepeat count: %d\n"),repeat_count);
-      return;
+      s->Show(Status());
+      current->TimeoutS(1);
    }
 }
 
@@ -214,7 +238,7 @@ Job *cmd_repeat(CmdExec *parent)
 }
 
 extern "C" {
-#include "getdate.h"
+#include "parse-datetime.h"
 }
 Job *cmd_at(CmdExec *parent)
 {
@@ -239,7 +263,7 @@ Job *cmd_at(CmdExec *parent)
    date.truncate(date_len);
 
    struct timespec ts;
-   if(!get_date(&ts,date,0))
+   if(!parse_datetime(&ts,date,0))
    {
       eprintf("%s: %s\n",args->a0(),"date parse error");
       return 0;
