@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: SleepJob.h,v 1.17 2009/03/18 11:30:20 lav Exp $ */
+/* $Id: SleepJob.h,v 1.18 2011/04/29 04:58:28 lav Exp $ */
 
 #ifndef SLEEPJOB_H
 #define SLEEPJOB_H
@@ -51,7 +51,7 @@ public:
    ~SleepJob();
 
    const char *Status();
-   void PrintStatus(int v,const char *);
+   xstring& FormatStatus(xstring&,int v,const char *);
    void ShowRunStatus(const SMTaskRef<StatusLine>& s);
 
    void Repeat(int m) { repeat=true; max_repeat_count=m; Stop(); }
