@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2010 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2011 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: commands.cc,v 1.299 2010/04/29 11:00:42 lav Exp $ */
+/* $Id: commands.cc,v 1.307 2011/04/29 04:58:29 lav Exp $ */
 
 #include <config.h>
 
@@ -86,8 +86,8 @@ CMD(exit);  CMD(get);    CMD(help);    CMD(jobs);
 CMD(kill);  CMD(lcd);    CMD(ls);      CMD(cls);
 CMD(open);  CMD(pwd);    CMD(set);     CMD(eval);
 CMD(shell); CMD(source); CMD(user);    CMD(rm);
-CMD(wait);  CMD(subsh);  CMD(mirror);
-CMD(mv);    CMD(cat);    CMD(cache);
+CMD(wait);  CMD(subsh);  CMD(mirror);  CMD(local);
+CMD(mv);    CMD(cat);    CMD(cache);   CMD(ln);
 CMD(mkdir); CMD(scache); CMD(mrm);
 CMD(ver);   CMD(close);  CMD(bookmark);CMD(lftp);
 CMD(echo);  CMD(suspend);CMD(sleep);   CMD(slot);
@@ -285,7 +285,10 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 " -u <user>[,<pass>]  use the user/password for authentication\n"
 	 " -p <port>           use the port for connection\n"
 	 " <site>              host name, URL or bookmark name\n")},
+   {"ln",      cmd_ln,	    N_("ln [-s] <file1> <file2>"),
+	 N_("Link <file1> to <file2>\n")},
    {"lpwd",    cmd_lpwd},
+   {"local",   cmd_local},
    {"login",   cmd_user,   0,"user"},
    {"ls",      cmd_ls,	    N_("ls [<args>]"),
 	 N_("List remote files. You can redirect output of this command to file\n"
@@ -424,8 +427,11 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
    {"repeat",  cmd_repeat, N_("repeat [OPTS] [delay] [command]"),
 	 N_("Repeat specified command with a delay between iterations.\n"
 	 "Default delay is one second, default command is empty.\n"
-	 " -c <count>  number of iterations\n"
-	 " -d <delay>  delay between iterations.\n")},
+	 " -c <count>  maximum number of iterations\n"
+	 " -d <delay>  delay between iterations\n"
+	 " --while-ok  stop when command exits with non-zero code\n"
+	 " --until-ok  stop when command exits with zero code\n"
+	 " --weak      stop when lftp moves to background.\n")},
    {"reput",   cmd_get,    0,
 	 N_("Usage: reput <lfile> [-o <rfile>]\n"
 	 "Same as `put -c'\n")},
@@ -446,7 +452,7 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
    	 " -a  list all settings, including default values\n"
 	 " -d  list only default values, not necessary current ones\n")},
    {"shell",   cmd_shell,  0,"!"},
-   {"site",    cmd_ls,	   N_("site <site_cmd>"),
+   {"site",    cmd_ls,	   N_("site <site-cmd>"),
 	 N_("Execute site command <site_cmd> and output the result\n"
 	 "You can redirect its output\n")},
    {"sleep",   cmd_sleep, 0,
@@ -461,7 +467,7 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
    {"source",  cmd_source, N_("source <file>"),
 	 N_("Execute commands recorded in file <file>\n")},
    {"suspend", cmd_suspend},
-   {"torrent", cmd_torrent, N_("torrent [-O <dir>] <file>")},
+   {"torrent", cmd_torrent, N_("torrent [-O <dir>] <file|URL>...")},
    {"user",    cmd_user,   N_("user <user|URL> [<pass>]"),
 	 N_("Use specified info for remote login. If you specify URL, the password\n"
 	 "will be cached for future usage.\n")},
@@ -596,6 +602,7 @@ Job *CmdExec::builtin_cd()
 	 return 0;
       }
       args->setarg(1,dir); // for status line
+      dir=args->getarg(1);
    }
 
    if(url::is_url(dir))
@@ -757,7 +764,9 @@ Job *CmdExec::builtin_lftp()
 	 cmd.set("version;");
 	 break;
       case('f'):
-	 cmd.vset("source \"",unquote(optarg),"\";",NULL);
+	 cmd.set("source ");
+	 cmd.append_quoted(optarg);
+	 cmd.append(';');
 	 break;
       case('c'):
 	 cmd.set_allocated(args->CombineCmd(args->getindex()-1));
@@ -897,31 +906,31 @@ Job *CmdExec::builtin_open()
 
    if(!no_bm && host && (bm=lftp_bookmarks.Lookup(host))!=0)
    {
-      char *cmd=string_alloca(5+3+3+xstrlen(user)*2+1+xstrlen(pass)*2+3+xstrlen(port)*2+strlen(bm)*2+6+1);
-      strcpy(cmd,"open -B ");
+      xstring& cmd=xstring::get_tmp("open -B ");
       if(user)
       {
-	 strcat(cmd,"-u \"");
-	 unquote(cmd+strlen(cmd),user);
+	 cmd.append("-u ");
+	 cmd.append_quoted(user);
 	 if(pass)
 	 {
-	    strcat(cmd,",");
-	    unquote(cmd+strlen(cmd),pass);
+	    cmd.append(",");
+	    cmd.append_quoted(pass);
 	 }
-	 strcat(cmd,"\" ");
+	 cmd.append(' ');
       }
       if(port)
       {
-	 strcat(cmd,"-p \"");
-	 unquote(cmd+strlen(cmd),port);
-	 strcat(cmd,"\" ");
+	 cmd.append("-p ");
+	 cmd.append_quoted(port);
+	 cmd.append(' ');
       }
 
-      strcat(cmd,bm);
+      cmd.append(bm);
 
       if(background)
-	 strcat(cmd," &\n");
-      strcat(cmd,";\n");
+	 cmd.append(" &\n");
+      else
+	 cmd.append(";\n");
 
       PrependCmd(cmd);
    }
@@ -937,9 +946,7 @@ Job *CmdExec::builtin_open()
 	    const char *p=ResMgr::Query("cmd:default-protocol",url->host);
 	    if(!p)
 	       p="ftp";
-	    char *u=string_alloca(strlen(p)+3+strlen(host)+1);
-	    sprintf(u,"%s://%s",p,host);
-	    url=new ParsedURL(u);
+	    url=new ParsedURL(xstring::format("%s://%s",p,host));
 	 }
 	 if(user || port)
 	 {
@@ -1056,13 +1063,11 @@ Job *CmdExec::builtin_open()
       }
 
       const char *cd_arg=(url && url->orig_url)?url->orig_url.get():path;
-      char *s=string_alloca(strlen(cd_arg)*2+40);
-      strcpy(s,"&& cd \"");
-      unquote(s+strlen(s),cd_arg);
-      strcat(s,"\"");
+      xstring& s=xstring::get_tmp("&& cd ");
+      s.append_quoted(cd_arg);
       if(background)
-	 strcat(s,"&");
-      strcat(s,"\n");
+	 s.append('&');
+      s.append('\n');
       PrependCmd(s);
    }
 
@@ -1132,6 +1137,7 @@ Job *CmdExec::builtin_glob()
       return cmd_command(this);
    }
    glob=new GlobURL(session,pat,glob_type);
+   RevertToSavedSession();
    builtin=BUILTIN_GLOB;
    return this;
 }
@@ -1223,7 +1229,11 @@ Job *CmdExec::builtin_queue()
 	       queue->Suspend();
 	    }
 	    else
-	       queue->PrintStatus(2,"");
+	    {
+	       xstring buf("");
+	       queue->FormatStatus(buf,2,"");
+	       printf("%s",buf.get());
+	    }
 	    exit_code=0;
 	    break;
 	 }
@@ -1305,53 +1315,6 @@ Job *CmdExec::builtin_queue()
    return 0;
 }
 
-#if 0
-Job *CmdExec::builtin_queue_edit()
-{
-   CmdExec *queue=GetQueue();
-
-   if(queue->feeder) {
-      eprintf(_("Can only edit plain queues.\n"));
-      return 0;
-   }
-
-   const char *home=getenv("HOME");
-   if(home==0)
-      home="";
-   /* if we have $HOME, use $HOME/.lftp/file; otherwise just file */
-   xstring_ca q_file(xasprintf("%s%slftp-queue-%i", home, home? "/.lftp/":"", (int) getpid()));
-
-   int q_fd=open(q_file,O_RDWR|O_CREAT,0600);
-   if(q_fd==-1) {
-      eprintf(_("Couldn't create temporary file `%s': %s.\n"), q_file.get(), strerror(errno));
-      return 0;
-   }
-
-   if(!queue->WriteCmds(q_fd)) {
-      eprintf(_("%s: error writing %s: %s\n"), args->a0(), q_file.get(), strerror(errno));
-
-      /* abort without clearing the queue */
-      close(q_fd);
-
-      return 0;
-   }
-
-   queue->EmptyCmds();
-
-   close(q_fd);
-
-   /* empty the queue; when this command exits, it'll re-source the queue.
-    * this also effectively "suspends" this queue until it's done editing. */
-
-   {
-      xstring_ca editcmd(xasprintf("shell \"/bin/sh -c 'exec ${EDITOR:-vi} %s'\"; queue -s %s; shell rm '%s'\n", q_file.get(), q_file.get(), q_file.get()));
-      PrependCmd(editcmd);
-   }
-
-   return 0;
-}
-#endif
-
 // below are only non-builtin commands
 #define args	  (parent->args)
 #define exit_code (parent->exit_code)
@@ -1964,9 +1927,35 @@ CMD(jobs)
 	 return 0;
       }
    }
-   parent->ListJobs(v);
    exit_code=0;
-   return 0;
+   args->back();
+   const char *op=args->a0();
+   const char *arg=args->getnext();
+   xstring s("");
+   if(!arg) {
+      parent->FormatJobs(s,v);
+   } else {
+      for(; arg; arg=args->getnext()) {
+	 if(!isdigit((unsigned char)*arg)) {
+	    eprintf(_("%s: %s - not a number\n"),op,arg);
+	    exit_code=1;
+	    continue;
+	 }
+	 int n=atoi(arg);
+	 Job *j=parent->FindJob(n);
+	 if(!j) {
+	    eprintf(_("%s: %d - no such job\n"),op,n);
+	    exit_code=1;
+	    continue;
+	 }
+	 j->FormatOneJob(s,v);
+      }
+   }
+   if(exit_code)
+      return 0;
+   OutputJob *out=new OutputJob(output.borrow(), args->a0());
+   Job *j=new echoJob(s,s.length(),out);
+   return j;
 }
 
 CMD(cd)
@@ -2369,7 +2358,7 @@ CMD(wait)
 
 CMD(subsh)
 {
-   CmdExec *e=new CmdExec(session->Clone(),parent->cwd->Clone());
+   CmdExec *e=new CmdExec(parent);
 
    const char *c=args->getarg(1);
    e->FeedCmd(c);
@@ -2390,6 +2379,33 @@ CMD(mv)
    return j;
 }
 
+CMD(ln)
+{
+   FA::open_mode m=FA::LINK;
+   const char *op=args->a0();
+   int c;
+   while((c=args->getopt("+s"))!=EOF)
+   {
+      switch(c)
+      {
+      case 's':
+	 m=FA::SYMLINK;
+	 break;
+      default:
+      error:
+	 eprintf(_("Try `help %s' for more information.\n"),op);
+	 return 0;
+      }
+   }
+   args->back();
+   const char *file1=args->getnext();
+   const char *file2=args->getnext();
+   if(!file1 || !file2)
+      goto error;
+
+   return new mvJob(session->Clone(),file1,file2,m);
+}
+
 const char *const cache_subcmd[]={
    "status","flush","on","off","size","expire",
    NULL
@@ -2584,7 +2600,7 @@ CMD(help)
 CMD(ver)
 {
    printf(
-      _("LFTP | Version %s | Copyright (c) 1996-%d Alexander V. Lukyanov\n"),VERSION,2010);
+      _("LFTP | Version %s | Copyright (c) 1996-%d Alexander V. Lukyanov\n"),VERSION,2011);
    printf("\n");
    printf(
       _("LFTP is free software: you can redistribute it and/or modify\n"
@@ -2622,7 +2638,6 @@ CMD(ver)
       typedef const char *(*func0)(int);
       const char *query() const
 	 {
-	    static char ver[10];
 	    int v;
 	    void *sym_ptr=dlsym(RTLD_DEFAULT,symbol);
 	    if(!sym_ptr)
@@ -2638,8 +2653,7 @@ CMD(ver)
 	       break;
 	    case INT8_8:
 	       v=*(int*)sym_ptr;
-	       sprintf(ver,"%d.%d",(v>>8)&255,v&255);
-	       str=ver;
+	       str=xstring::format("%d.%d",(v>>8)&255,v&255);
 	    }
 	    if(!str)
 	       return 0;
@@ -2666,7 +2680,7 @@ CMD(ver)
       if(!v)
 	 continue;
       char buf[256];
-      sprintf(buf,", %s %s",scan->lib_name,v);
+      snprintf(buf,sizeof(buf),", %s %s",scan->lib_name,v);
       int skip=need_comma?0:2;
       int w=mbswidth(buf+skip,mbflags);
       if(col+w>=width)
@@ -2759,11 +2773,10 @@ CMD(bookmark)
 	 if(value==0)
 	 {
 	    value=session->GetConnectURL(flags);
-	    char *a=string_alloca(strlen(value)*3+2);
 	    // encode some more characters, special to CmdExec parser.
-	    url::encode_string(value,a,"&;|\"'\\");
+	    xstring& a=url::encode(value,"&;|\"'\\");
 	    if(value[0] && last_char(value)!='/')
-	       strcat(a,"/");
+	       a.append('/');
 	    value=a;
 	 }
 	 if(value==0 || value[0]==0)
@@ -2802,10 +2815,7 @@ CMD(bookmark)
 	 eprintf(_("%s: import type required (netscape,ncftp)\n"),args->a0());
       else
       {
-	 const char *fmt="shell " PKGDATADIR "/import-%s\n";
-	 char *cmd=string_alloca(strlen(op)+strlen(fmt)+1);
-	 sprintf(cmd,fmt,op);
-	 parent->PrependCmd(cmd);
+	 parent->PrependCmd(xstring::cat("shell " PKGDATADIR "/import-",op,"\n",NULL));
 	 exit_code=0;
       }
    }
@@ -3079,12 +3089,16 @@ CMD(lpwd)
       return 0;
    }
    const char *name=parent->cwd->GetName();
-   char *buf=alloca_strdup2(name,2);
-   sprintf(buf,"%s\n",name?name:"?");
+   const char *buf=xstring::cat(name?name:"?","\n",NULL);
    Job *j=new echoJob(buf,new OutputJob(output.borrow(), args->a0()));
    return j;
 }
 
+CMD(local)
+{
+   return parent->builtin_local();
+}
+
 CMD(glob)
 {
    return parent->builtin_glob();
@@ -3252,10 +3266,7 @@ CMD(get1)
 	 const char *bn=basename_ptr(src);
 	 if(bn[0]!='/')
 	 {
-	    char *dst1=string_alloca(strlen(dst)+strlen(bn)+1);
-	    strcpy(dst1,dst);
-	    strcat(dst1,bn);
-	    dst=dst1;
+	    dst=xstring::get_tmp(dst).append(bn);
 	    auto_rename=true;
 	 }
       }
@@ -3277,11 +3288,7 @@ CMD(get1)
 	       slash++;
 	    else
 	       slash=src;
-	    char *dst1=string_alloca(strlen(dst)+strlen(slash)+2);
-	    strcpy(dst1,dst);
-	    strcat(dst1,"/");
-	    strcat(dst1,slash);
-	    dst=dst1;
+	    dst=xstring::cat(dst,"/",slash,NULL);
 	 }
       }
    }
