@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2010 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -47,6 +47,7 @@
 #endif
 
 #include <stdarg.h>
+#include <unistd.h>
 
 #define string_alloca(len)      ((char*)alloca((len)))
 
@@ -79,6 +80,8 @@ protected:
    char *buf;
    xstring0() {}
    xstring0(const xstring0&); // disable cloning
+
+   int _url_decode(size_t len);
 public:
    ~xstring0() { xfree(buf); }
    operator const char *() const { return buf; }
@@ -103,15 +106,18 @@ public:
    const char *set(const char *s) { return xstrset(buf,s); }
    const char *nset(const char *s,int n) { return xstrset(buf,s,n); }
    const char *set_allocated(char *s) { xfree(buf); return buf=s; }
-   const char *vset(...) __attribute__((sentinel));
-   void truncate(size_t n) { if(buf) buf[n]=0; }
+   const char *vset(...) ATTRIBUTE_SENTINEL;
+   void truncate(size_t n=0) { if(buf) buf[n]=0; }
    char *borrow() { return replace_value(buf,(char*)0); }
    bool eq(const char *s) { return !xstrcmp(buf,s); }
    bool ne(const char *s) { return !eq(s); }
    size_t length() const { return xstrlen(buf); }
+   void set_length(size_t n) { if(buf) buf[n]=0; }
 
    void unset() { xfree(buf); buf=0; }
    void _set(const char *s) { buf=xstrdup(s); }
+
+   xstring_c& url_decode();
 };
 class xstring_ca : public xstring_c
 {
@@ -142,9 +148,9 @@ public:
    xstring(const char *s) { init(s); }
    xstring(const char *s,int l) { init(s,l); }
 
-   // allocates s bytes, with preferred granularity g
-   void get_space(size_t s,size_t g=32);
-   char *add_space(size_t s,size_t g=32) { get_space(len+s,g); return get_non_const()+len; }
+   void get_space(size_t s);
+   void get_space2(size_t s,size_t g);
+   char *add_space(size_t s);
    void add_commit(int new_len) { len+=new_len; }
 
    size_t length() const { return len; }
@@ -163,8 +169,8 @@ public:
    xstring& append(const xstring &s) { return append(s.get(),s.length()); }
    xstring& append_padding(int len,char ch);
    xstring& vappend(va_list);
-   xstring& vappend(...) __attribute__((sentinel));
-   xstring& vset(...) __attribute__((sentinel));
+   xstring& vappend(...) ATTRIBUTE_SENTINEL;
+   xstring& vset(...) ATTRIBUTE_SENTINEL;
    xstring& vsetf(const char *fmt, va_list ap) { truncate(0); return vappendf(fmt,ap); }
    xstring& setf(const char *fmt,...) PRINTF_LIKE(2,3);
    xstring& vappendf(const char *fmt, va_list ap);
@@ -174,10 +180,10 @@ public:
    static xstring& get_tmp(const char *s,int n) { return get_tmp().nset(s,n); }
    static xstring& vformat(const char *fmt,va_list ap) { return get_tmp().vsetf(fmt,ap); }
    static xstring& format(const char *fmt,...) PRINTF_LIKE(1,2);
-   static xstring& cat(const char *first,...) __attribute__((sentinel));
+   static xstring& cat(const char *first,...) ATTRIBUTE_SENTINEL;
    static xstring& join(const char *sep,int n,...);
 
-   void truncate(size_t n);
+   void truncate(size_t n=0);
    void truncate_at(char c);
    /* set_length can be used to extend the string, e.g. after modification
       with get_space+get_non_const. */
@@ -200,7 +206,18 @@ public:
    const char *dump_to(xstring &out) const;
    const char *dump() const;
 
+   xstring& url_decode();
+   xstring& append_url_encoded(const char *s,int len,const char *unsafe);
+   xstring& append_url_encoded(const char *s,const char *unsafe) { return append_url_encoded(s,strlen(s),unsafe); }
+   xstring& append_url_encoded(const xstring& s,const char *unsafe) { return append_url_encoded(s,s.length(),unsafe); }
+
+   xstring& append_quoted(const char *s,int len);
+   xstring& append_quoted(const char *s) { return append_quoted(s,strlen(s)); }
+
    static xstring null;
 };
 
+static inline size_t strlen(const xstring& s) { return s.length(); }
+static inline size_t xstrlen(const xstring& s) { return s.length(); }
+
 #endif//XSTRING_H
